'''
Created on 2013-07-02

@author: Jackie
sigtest.py - functions for doing statistical significance tests
'''

import sys
from scipy.stats import wilcoxon
from sklearn.metrics import auc_score
import numpy as np
import random


ntrials = 100000 # eventually, 10^6
def bootstrap_sigtest(c1_f, c2_f):
    '''
    Task 1
    Run the subsampling based method of comparing two conditions on the same
    dataset. Refer to Berg-Kirkpatrick et. al, 2012 for details.
    '''
    # read details_f
    c1 = read_task1_details(c1_f)
    c2 = read_task1_details(c2_f)
    
    # get original difference in avg. auc
    auc1 = avg_auc(c1)
    auc2 = avg_auc(c2)
    d_auc = auc2 - auc1
    # positive: c2 better than c1
    # negative: c1 better than c2
    print 'C1 AUC:      %.4f' % auc1
    print 'C2 AUC:      %.4f' % auc2
    print 'Original d:  %.4f' % d_auc
    
    # do random sampling
    nsig = 0
    for i in xrange(ntrials):
        if i % 100 == 0: sys.stderr.write('  iteration %d...\n' % i)
        # sample the entities WITH REPLACEMENT
        d_auc_i = sample_difference(c1, c2)
        
        # tally whether it is twice as great for p, r, f1
        signif = int(d_auc == 0.0 or d_auc_i / d_auc > 2.0)
        nsig += signif
        #if signif == 1:
        #    print 'Sample %d d:  %.4f*' % (i, d_auc_i)
        #else:
        #    print 'Sample %d d:  %.4f' % (i, d_auc_i)

    pval = float(nsig) / ntrials
    # output results, correcting for bias
    print 'AUC     p = %d / %d = %.4f' % (nsig, ntrials, pval)

def read_task1_details(fname):
    data = {}
    lines = open(fname).read().split('\n')
    if lines[-1] == '': lines = lines[:-1]
    
    for i in xrange(0, len(lines), 3):
        data[lines[i]] = (eval(lines[i+1]), eval(lines[i+2]))
    return data

def avg_auc(data):
    '''
    Calculates AUC for each subset in data and returns the average.
    '''
    aucs = []
    for k, (labels, scores) in data.iteritems():
        aucs.append(auc_score(labels, scores))
    return np.mean(np.array(aucs))

def sample_difference(c1, c2):
    '''
    Do sampling from each subset, calculate the AUCs, and return the difference. 
    '''
    s1, s2 = {}, {}
    subsets = sorted(c1.keys())
    assert subsets == sorted(c2.keys())
    
    for subset in subsets:
        s1[subset] = [], []
        s2[subset] = [], []
        labels1, scores1 = c1[subset]
        labels2, scores2 = c2[subset]
        assert labels1 == labels2
        for n in xrange(len(scores1)):
            # sample an equal-sized sample with replacement
            r = random.randint(0, len(scores1) - 1)
            s1[subset][0].append(labels1[r])
            s2[subset][0].append(labels2[r])
            s1[subset][1].append(scores1[r])
            s2[subset][1].append(scores2[r])
    
    try:
        auc1 = avg_auc(s1)
        auc2 = avg_auc(s2)
    except:
        # sometimes, it'll fail because it sampled only from one class---repeat
        return sample_difference(c1, c2)
    return auc2 - auc1
        
def read_task2_results(fname):
    fh = open(fname)
    all_ranks = eval(fh.readline()[:-1])
    all_mr = eval(fh.readline()[:-1])
    subset_ranks = eval(fh.readline()[:-1])
    subset_mr = eval(fh.readline()[:-1])
    return all_ranks, all_mr, subset_ranks, subset_mr
    
def paired_wilcoxon_sigtest(c1_f, c2_f):
    '''
    Do a paired Wilcoxon signed rank test.
    '''
    c1_all, _, c1_subset, _ = read_task2_results(c1_f)
    c2_all, _, c2_subset, _ = read_task2_results(c2_f)
    
    assert len(c1_all) == len(c2_all) 
    print 'Full set N = %d' % len(c1_all)
    print wilcoxon(c1_all, c2_all)
    
    assert len(c1_subset) == len(c2_subset) 
    print 'Subset N = %d' % len(c1_subset)
    print wilcoxon(c1_subset, c2_subset)
     
    