'''
Created on 2013-04-29

@author: Jackie
'''
from dseval.data_structures import Document, Sentence, Token, CorefChain, CorefMention, DepNode
from xml.etree import cElementTree as ElementTree
import os, sys
import getopt


def read_dict(f, splitf = lambda x: x.strip().split('\t'), kf = None, vf = None, commentf = None):
    '''
    Read a tab delimited kf(str) -> vf(str) dict.
    Pass in None to kf or vf to keep the original str.
    Removes leading and trailing whitespace.
    commentf is a function that returns True if a line is a comment
    '''
    if isinstance(f, str):
        f = open(f)
    d = {}
    for line in f.readlines():
        if len(line) <= 1:
            continue
        if commentf is not None and commentf(line):
            continue
        args = splitf(line)
        k = args[0]
        if kf is not None:
            k = kf(k)
        try:
            v = args[1]
        except:
            print args
            continue
        if vf is not None:
            v = vf(v)
        d[k] = v
    return d


def read_list(f, vf = None):
    '''
    Read a list where each line is a separate entry.
    Removes leading and trailing whitespace.
    '''
    if isinstance(f, str):
        f = open(f)
    lst = []
    for line in f.readlines():
        if len(line) <= 1:
            continue
        v = line.strip()
        if vf is not None:
            v = vf(v)
        lst.append(v)
    return lst


def write_dict(towrite, f, kf = None, vf = None, close = True):
    '''
    Writes a dict to file
    '''
    if isinstance(f, str):
        f = open(f, 'w')
    for k, v in towrite.iteritems():
        if kf is not None:
            k = kf(k)
        if vf is not None:
            v = vf(v)
        f.write('%s\t%s\n' % (k, v))
    if close:
        f.close()
        

def write_list(towrite, f, vf = None, close = True):
    '''
    Writes a list to file
    '''
    if isinstance(f, str):
        f = open(f, 'w')
    for el in towrite:
        if vf is not None:
            el = vf(el)
        f.write('%s\n' % el)
    if close:
        f.close()
        
def read_corenlp_corpus(directory, filedict, toload = None):
    '''
    A generator of (docnum, Document) tuples.
    directory - folder containing the CoreNLP output files
    filedict - docnum <int> -> filename (relative to directory) <str>
    
    Can use either the keys of filedict (default) or pass in a
    particular list of keys to load. 
    '''
    if toload is None:
        keys = filedict.keys()
        keys.sort()
    else:
        keys = sorted(toload)
        
    for docnum in keys:
        filename = filedict[docnum]
        doc = read_corenlp_doc(os.path.join(directory, filename), docnum)
        yield (docnum, doc)
        
def read_corenlp_doc(xmlf, docnum):
    '''
    New version that uses ElementTree
    '''
    tree = ElementTree.parse(xmlf)
    docnode = tree.find('document')
    doc = read_doc_from_node(docnode)
    doc.num = docnum
    return doc

def read_doc_from_node(docnode, corenlp_format = True, treetype = 'collapsed-ccprocessed-dependencies'):
    '''
    corenlp_format
      = True: <governor idx="#">word</governor>
      = False (annotated_gigaword_format): <governor>#</governor>
      
    treetype
      = basic-dependencies, collapsed-dependencies, or collapsed-ccprocessed-dependencies
    '''
    doc = Document()
    
    sents_node = docnode.find('sentences')
    if sents_node is None: # no sentences
        return doc
    sentnodes = sents_node.findall('sentence')
    for sentnode in sentnodes:
        # load basic information about tokens in sentences
        sent = Sentence()
        sent.num = int(sentnode.get('id'))
        doc.sents[sent.num] = sent
        tokennodes = sentnode.find('tokens').findall('token')
        for tokennode in tokennodes:
            token = Token()
            token.num = int(tokennode.get('id'))
            token.word = tokennode.findtext('word')
            token.lemma = tokennode.findtext('lemma') 
            token.pos = tokennode.findtext('POS')
            token.ner = tokennode.findtext('NER')
            sent.tokens[token.num] = token
        
        # load parse
        depnodes = sentnode.find(treetype).findall('dep')
        for depnode in depnodes:
            node = DepNode()
            node.reltype = depnode.get('type')
            if corenlp_format:
                node.governor = int(depnode.find('governor').get('idx'))
                node.dependent = int(depnode.find('dependent').get('idx'))
            else:
                node.governor = int(depnode.findtext('governor'))
                node.dependent = int(depnode.findtext('dependent'))
            sent.parse.add(node.dependent, node)
        __add_head(sent)
    # load coreference chains for document
    corefstring = 'coreference'
    if not corenlp_format: corefstring = 'coreferences' 
    corefs_node = docnode.find(corefstring)
    if corefs_node is not None:
        for corefnode in corefs_node.getchildren():
            chain = CorefChain()
            for mentionnode in corefnode.findall('mention'):
                mention = CorefMention()
                mention.representative = 'representative' in mentionnode.keys()
                mention.sentence = int(mentionnode.findtext('sentence'))
                mention.start = int(mentionnode.findtext('start'))
                mention.end = int(mentionnode.findtext('end'))
                mention.head = int(mentionnode.findtext('head'))
                chain.mentions.append(mention)
            if len(chain.mentions) > 0:
                doc.corefs.append(chain)
    return doc

    
def __add_head(sent):
    '''
    Add the head of sent as a node in sent.parse, because
    Stanford parser does not do this automatically.
    
    Head of sentence: has dependents AND no governors
    Add it back in with governor = 0
    '''
    for tokennum in sent.tokens:
        if not sent.parse.nodes.has_key(tokennum) and \
           len(sent.parse.dependents_of(tokennum)) > 0:
            headnode = DepNode()
            headnode.reltype = 'root'
            headnode.governor = 0
            headnode.dependent = tokennum
            sent.parse.add(tokennum, headnode)
            break 
        
        
def cmdline_io(acceptable_opts, help_func, default_config = None):
    '''Reads basic command line options.'''
    params = {} 
    if default_config is not None:
        params['-c'] = default_config

    # read in parameters
    opts, args = getopt.getopt(sys.argv[1:], acceptable_opts)
    for o, a in opts:
        params[o] = a

    if params.has_key('-h'):
        help_func()
    
    # print parameters
    print >> sys.stderr, 'Parameters:'
    for param, val in params.iteritems():
        print >> sys.stderr,  '%s\t%s' % (param, val)
    print >> sys.stderr
    
    # read and print configuration parameters
    commentf = lambda x: len(x.strip()) == 0 or x.strip().startswith('#')
    c = {}
    if params.has_key('-c'): 
        c = read_dict(params['-c'], commentf = commentf)
        root_d = c.get('root_d', '')
        #print >> sys.stderr,  'Configuration:'
        for param in c.keys():
            val = c[param]
            if not val.startswith('/') or (len(val) > 2 and val[1] != ':'):
                c[param] = os.path.join(root_d, val)
            #print >> sys.stderr,  '%s\t%s' % (param, c[param])
        #print >> sys.stderr
    
    return params, c